#include <algorithm>
#include <assert.h>
#include "ztree.h"
#include <iomanip>
#include "logger.h"
#include "glhelper/baseglutils.h"

using namespace math3;

int bit_interlace_table[2048];
class ztree_initializer{
	public:
	ztree_initializer(){
		for(int i=0;i<2048;++i){
			int v=0;
			for(int j=0;j<11;++j){
				v|=((i>>j)&1)<<(j*3);
			}
			bit_interlace_table[i]=v;
		}
	}
};
static ztree_initializer initializer;
int64_t BitInterlace_11(int x, int y, int z){
	x&=2047;
	y&=2047;
	z&=2047;
	return (((int64_t(bit_interlace_table[z])<<1)|bit_interlace_table[y])<<1)|bit_interlace_table[x];
}
#ifdef BUILDTYPE_PROFILE
int64_t BitInterlace_21(int x, int y, int z) __attribute__((noinline));
#endif
int64_t BitInterlace_21(int x, int y, int z){
	return (BitInterlace_11(x>>11,y>>11,z>>11)<<33)|BitInterlace_11(x,y,z);
}

void BitDeinterlace(int64_t v, int &x, int &y, int &z){
	x=0;
	y=0;
	z=0;
	int i=0;
	while(v){
		x|=(v&1)<<i;
		y|=(v&2)<<i;
		z|=(v&4)<<i;
		v>>=3;
		i++;
	}
	y>>=1;
	z>>=2;
}


Collisions::Collisions():
last_collision_id(-1), raw_collision_count(0)
{
}

void Collisions::Clear(){
	ids.clear();
	last_collision_id=-1;
}

void Collisions::MakeCollisionsUnique(){
	std::sort(ids.begin(),ids.end());
	auto new_end=std::unique(ids.begin(),ids.end());
	ids.resize(new_end-ids.begin());
}

CollisionDetector::CollisionDetector():scale(100),offset(0.5*resolution, 0.5*resolution, 0.5*resolution){
}
void CollisionDetector::Clear(){
	entries.clear();
}
const int invalid_contained_by=0xffffff;
void AddPoint_raw(std::vector<CollisionDetector::Entry> &entries, int id, const math3::Vec3<int> &pos){
	CollisionDetector::Entry entry;
	entry.start=BitInterlace_21(pos.x, pos.y, pos.z);
	entry.size=0;
	entry.contained_by=invalid_contained_by;
	entry.object_id=id;
	entries.push_back(entry);
}
void AddAABB_raw(std::vector<CollisionDetector::Entry> &entries, int id, const math3::Vec3<int> &pos, const math3::Vec3<int> &size){
	int d=size.x;
	if(d<size.y)d=size.y;
	if(d<size.z)d=size.z;
	int sz2;/// cell size, power of 2.
	int szp;/// actual power of 2
	for(sz2=1, szp=0; sz2<d; sz2<<=1, szp++);
	int lx=pos.x;
	int ly=pos.y;
	int lz=pos.z;
	if(lx<0)lx=0;
	if(ly<0)ly=0;
	if(lz<0)lz=0;
	int hx=pos.x+size.x;
	int hy=pos.y+size.y;
	int hz=pos.z+size.z;
	if(hx>=CollisionDetector::resolution)hx=CollisionDetector::resolution-1;
	if(hy>=CollisionDetector::resolution)hy=CollisionDetector::resolution-1;
	if(hz>=CollisionDetector::resolution)hz=CollisionDetector::resolution-1;
	int mask=sz2-1;
	lx&=~mask;
	ly&=~mask;
	lz&=~mask;
	hx|=mask;
	hy|=mask;
	hz|=mask;
	int iterations=0;
	if(szp>10){
		//std::cout<<"wtf"<<std::endl;
	}
	for(int iz=lz; iz<=hz ; iz+=sz2){
		for(int iy=ly; iy<=hy; iy+=sz2){
			for(int ix=lx; ix<=hx; ix+=sz2){
				///should be max. 8 iterations.
				CollisionDetector::Entry entry;
				entry.start=BitInterlace_21(ix, iy, iz);
				entry.size=szp;
				entry.contained_by=invalid_contained_by;
				entry.object_id=id;
				entries.push_back(entry);
				iterations++;
			}
		}
	}
#ifdef BUILDTYPE_DEBUG
	if(iterations>8){
		std::cout<<"too many iterations"<<std::endl;
	}
#endif
}
void AddSphere_raw(std::vector<CollisionDetector::Entry> &entries, int id, const math3::Vec3d &pos, float r){
	Vec3<int> l(pos.x-r,pos.y-r,pos.z-r);
	Vec3<int> h(pos.x+r+1.0,pos.y+r+1.0,pos.z+r+1.0);
	AddAABB_raw(entries, id, l, h-l);
}
void CollisionDetector::AddPoint(int id, const math3::Vec3f &pos){
	AddPoint_raw(entries, id, Vec3<int>(Vec3d(pos)*scale+offset));
}
void CollisionDetector::AddAABB(int id, const math3::Vec3f &pos, const math3::Vec3f &size){
	AddAABB_raw(entries, id, Vec3d(pos)*scale+offset, size*float(scale));
}
void CollisionDetector::AddAABB(int id, const Box3f &box){
	AddAABB(id, box.low, box.high-box.low);
}
void CollisionDetector::AddSphere(int id, const math3::Vec3f &pos, float r){
	AddSphere_raw(entries, id, Vec3d(pos)*scale+offset, r*scale);
}
void CollisionDetector::AddSphere(int id, const math3::Vec3f &pos1, const math3::Vec3f &pos2, float r){
	Box3<float> box(pos1,pos2);
	box.Sweep(r);
	AddAABB(id,box);
}
void CollisionDetector::Process(Collisions &result){
	std::sort(entries.begin(),entries.end());
	for(int i=0;i<entries.size();++i){
		int j=i+1;
		int64_t range_end=entries[i].range_end();
		while(j<entries.size() && entries[j].start<range_end){
			Entry &a=entries[i];
			Entry &b=entries[j];
			if(a.object_id==b.object_id){
				logger<<"CollisionDetector bug"<<std::endl;
			}
			/// uncomment this line to disable reporting of collision for identical boxes
			//if(a.size>b.size)
			b.contained_by=i;
			result.AppendCollision(Collisions::CollisionID(a.object_id,b.object_id));
			j++;
		}
	}
	result.raw_collision_count=result.ids.size();
}


void CollisionDetector::TestAgainst(const CollisionDetector &other, Collisions &result){
	for(int i=0;i<other.entries.size();++i){
		const Entry &other_entry=other.entries[i];
		int j=(std::upper_bound(entries.begin(),entries.end(), other_entry)-entries.begin());
		/// collide with entries which may be contained by other entry (walk down the tree):
		int64_t range_end=other_entry.range_end();
		for(int k=j; k<entries.size() && entries[k].start<range_end; ++k){
			result.AppendCollision(Collisions::CollisionID(other_entry.object_id, entries[j].object_id));
		}
		/// collide with entries which may contain other entry (walk up the tree):
		j-=1;
		while(j>=0 && j!=invalid_contained_by){
			int64_t range_end=entries[j].range_end();
			if(range_end>other_entry.start){
				result.AppendCollision(Collisions::CollisionID(other_entry.object_id, entries[j].object_id));
			}
			j=entries[j].contained_by;
		}
	}
	result.raw_collision_count=result.ids.size();
}

void DrawBox(std::vector<math3::Vec3f> &pts, const math3::Vec3f &a, const math3::Vec3f &b){
	pts.push_back(Vec3f(a.x,a.y,a.z));
	pts.push_back(Vec3f(b.x,a.y,a.z));

	pts.push_back(Vec3f(b.x,a.y,a.z));
	pts.push_back(Vec3f(b.x,b.y,a.z));

	pts.push_back(Vec3f(b.x,b.y,a.z));
	pts.push_back(Vec3f(a.x,b.y,a.z));

	pts.push_back(Vec3f(a.x,b.y,a.z));
	pts.push_back(Vec3f(a.x,a.y,a.z));


	pts.push_back(Vec3f(a.x,a.y,b.z));
	pts.push_back(Vec3f(b.x,a.y,b.z));

	pts.push_back(Vec3f(b.x,a.y,b.z));
	pts.push_back(Vec3f(b.x,b.y,b.z));

	pts.push_back(Vec3f(b.x,b.y,b.z));
	pts.push_back(Vec3f(a.x,b.y,b.z));

	pts.push_back(Vec3f(a.x,b.y,b.z));
	pts.push_back(Vec3f(a.x,a.y,b.z));


	pts.push_back(Vec3f(a.x,a.y,a.z));
	pts.push_back(Vec3f(a.x,a.y,b.z));

	pts.push_back(Vec3f(b.x,a.y,a.z));
	pts.push_back(Vec3f(b.x,a.y,b.z));

	pts.push_back(Vec3f(b.x,b.y,a.z));
	pts.push_back(Vec3f(b.x,b.y,b.z));

	pts.push_back(Vec3f(a.x,b.y,a.z));
	pts.push_back(Vec3f(a.x,b.y,b.z));
}

void CollisionDetector::DrawBoxes(int my_id){
	std::vector<Vec3f> vertices;
	//glBegin(GL_LINES);
	float a=1.0/scale;
	Vec3f b=-offset*double(a);
	for(int i=0;i<entries.size();++i){
		Entry &entry=entries[i];
		if(entry.object_id!=my_id){
			int x,y,z;
			BitDeinterlace(entry.start,x,y,z);
			int s=1<<entry.size;
			DrawBox(vertices,Vec3f(x,y,z)*a+b, Vec3f(x+s,y+s,z+s)*a+b);
		}
	}
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glEnableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glVertexPointer(3,GL_FLOAT,sizeof(Vec3f),(void*)&vertices[0]);
	glDrawArrays(GL_LINES,0,vertices.size());
	//glEnd();
}
