#ifndef ZTREE_H_
#define ZTREE_H_
#include <vector>
#include <math3/math3.h>
template<class T>
class Box3{
	public:
	math3::Vec3<T> low, high;
	Box3(const math3::Vec3<T> &low_, const math3::Vec3<T> &high_)
	{
		low.x =low_.x<high_.x?low_.x:high_.x;
		low.y =low_.y<high_.y?low_.y:high_.y;
		low.z =low_.z<high_.z?low_.z:high_.z;
		high.x=low_.x<high_.x?high_.x:low_.x;
		high.y=low_.y<high_.y?high_.y:low_.y;
		high.z=low_.z<high_.z?high_.z:low_.z;
	}
	template<class U>
	void Sweep(const Box3<U> &other){
		low+=other.low;
		high+=other.high;
	}
	template<class U>
	void Sweep(U other){
		low-=math3::Vec3<T>(other,other,other);
		high+=math3::Vec3<T>(other,other,other);
	}
};

typedef Box3<float> Box3f;

inline Box3f BoxFromSweptSphere(const math3::Vec3f &pos1, const math3::Vec3f &pos2, float radius){
	Box3f result(pos1,pos2);
	result.Sweep(radius);
	return result;
}


class Collisions{
	public:
	std::vector<int64_t> ids;// contains IDs of all collisions noted
	static inline int64_t CollisionID(int id_1, int id_2){
		return id_1<id_2 ? id_1|(int64_t(id_2)<<32) : id_2|(int64_t(id_1)<<32);
	}
	int64_t last_collision_id;/// used to cut down number of duplicates somewhat prior to sorting
	inline void AppendCollision(int64_t collision_id){
		if(last_collision_id!=collision_id){
			ids.push_back(collision_id);
			last_collision_id=collision_id;
		}
	}
	int raw_collision_count;

	Collisions();
	void MakeCollisionsUnique();
	void Clear();
};

/// Z-order based collision detection.
class CollisionDetector{
	public:
	enum {
		resolution=1<<21
	};
	struct Entry{
		/// start<=a<end
		int64_t start;
		unsigned contained_by:24;
		unsigned size:8;
		int object_id;
		inline int64_t range_end() const{
			return start+(int64_t(1)<<(size*3));
		}
		inline bool operator<(const Entry &other) const{/// ordering: z order, plus on equal sizes largest first
			return start==other.start? size>other.size: start<other.start;
		}
	};
	std::vector<Entry> entries;/// currently up to 16M entries will work correctly.
	double scale;
	math3::Vec3d offset;

	CollisionDetector();
	void Clear();
	void AddPoint(int id, const math3::Vec3f &pos);
	void AddAABB(int id, const math3::Vec3f &pos, const math3::Vec3f &size);
	void AddAABB(int id, const Box3f &box);

	void AddSphere(int id, const math3::Vec3f &pos, float r);
	/// swept sphere
	void AddSphere(int id, const math3::Vec3f &pos1, const math3::Vec3f &pos2, float r);

	/// after adding all items, call this to get collisions between items
	void Process(Collisions &result);

	/// Calculates collisions with other. Call this method after calling Process() in this instance.
	/// No need to call process in other instance
	void TestAgainst(const CollisionDetector &other, Collisions &result);
	void DrawBoxes(int my_id=-1);
};

#endif // ZTREE_H_
